package org.apollo.gui;

import java.util.LinkedList;

import org.apollo.ApolloGestureActions;
import org.apollo.audio.util.Timeline;
import org.apollo.items.FramePlaybackLauncher;
import org.apollo.items.RecordOverdubLauncher;
import org.apollo.util.Mutable;
import org.apollo.widgets.LinkedTrack;
import org.apollo.widgets.SampledTrack;
import org.expeditee.core.Clip;
import org.expeditee.core.Colour;
import org.expeditee.core.Dimension;
import org.expeditee.core.Fill;
import org.expeditee.core.Point;
import org.expeditee.core.Stroke;
import org.expeditee.core.TextLayout;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.gio.gesture.StandardGestureActions;
import org.expeditee.gio.input.KBMInputEvent.Key;
import org.expeditee.gio.input.StandardInputEventListeners;
import org.expeditee.gui.Browser;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FreeItems;
import org.expeditee.gui.FrameGraphics.FrameRenderPass;
import org.expeditee.items.Item;
import org.expeditee.items.widgets.Widget;
import org.expeditee.items.widgets.WidgetCorner;
import org.expeditee.items.widgets.WidgetEdge;

/**
 * All final effects rendered here - to give extra feedback to the user and simply to make
 * it more appealing.
 * @author Brook Novak
 *
 */
public class FrameRenderPasses implements FrameRenderPass {
	
	private static final Colour SHADED_EXPANDED_BACKING_COLOR = Colour.FromRGBA255(80, 80, 80, 140);
	private static final Colour Y_HELPER_LINES_COLOR = Colour.DARK_GREY;
	private static final Stroke Y_HELPER_LINES_STROKE = OverdubbedFrameTimeAxis.TRACK_TIMELINE_STROKE;
	
	private static FrameRenderPasses instance = new FrameRenderPasses();
	
	private LinkedList<Integer> timeHelperLinesToPaint = null;
	private Mutable.Integer yAxisTimelineHelperToPaint = null;
	
	private boolean shouldDrawFullscreenHelpers = false;
	
	private FrameRenderPasses()
	{
		FrameGraphics.addFrameRenderPass(this);
	}
	
	public static FrameRenderPasses getInstance()
	{
		return instance;
	}

	public void paintFinalPass()
	{
		if (Browser._theBrowser == null) return;
		
		GraphicsManager gm = EcosystemManager.getGraphicsManager();
		
		int frameHeight = EcosystemManager.getGraphicsManager().getWindowSize().getHeight();
		if (ExpandedTrackManager.getInstance().isAnyExpandedTrackVisible()
				&& !FreeItems.getInstance().isEmpty()) {

			// Conceivable could have multiple
			LinkedList<SampledTrack> toPaint = null;
			
			for (Item i : FreeItems.getInstance()) {
				if (i instanceof WidgetCorner) {
					Widget iw = ((WidgetCorner)i).getWidgetSource();
					
					if (iw != null && iw instanceof SampledTrack) {
						
						if (toPaint == null) {
							toPaint = new LinkedList<SampledTrack>();
						} else if (toPaint.contains(iw)) {
							continue;
						}
						
						toPaint.add((SampledTrack)iw);
					}
				}
			}
			
			if (toPaint != null) {
				for (SampledTrack trackWidget : toPaint) {
					if (gm.peekClip() == null || gm.peekClip().isNotClipped() || (!gm.peekClip().isFullyClipped() && gm.peekClip().getBounds().intersects(trackWidget.getBounds()))) {
						trackWidget.paintInFreeSpace(true);
					}
				}
			}
			
		} 
		
		OverdubbedFrameTimeAxis.getInstance().paint();
		
		TimeAxis timeAxis = OverdubbedFrameTimeAxis.getInstance().getTimeAxis();	
		Timeline lastTimeline = OverdubbedFrameTimeAxis.getInstance().getLastTimeline();
		
		FramePlaybackBarRenderer.getInstance().paint();

		if (timeHelperLinesToPaint != null && timeAxis != null && lastTimeline != null) {

			AxisAlignedBoxBounds r = OverdubbedFrameTimeAxis.getInstance().getCurrentTimelineArea();
			
			// Draw helper bar and text
			for (Integer x : timeHelperLinesToPaint) {
				
				if (shouldDrawFullscreenHelpers) {
					gm.drawLine(new Point(x, 0), new Point(x, frameHeight), Colour.RED, OverdubbedFrameTimeAxis.TRACK_TIMELINE_STROKE);
				} else {
					gm.drawLine(new Point(x, r.getMinY()), new Point(x, r.getMinY() + r.getHeight()), Colour.RED, OverdubbedFrameTimeAxis.TRACK_TIMELINE_STROKE);
				}
				
				long ms = lastTimeline.getMSTimeAtX(x) - lastTimeline.getFirstInitiationTime();

				String label = timeAxis.createSpecificTimeLabel(Math.abs(ms), 500.0);
				if (ms < 0) label = "-" + label;
				
				//FontMetrics fm   = g.getFontMetrics(SampledTrackGraphViewPort.TIME_HELPER_FONT);
				//Rectangle2D rect = fm.getStringBounds(label, g);
				TextLayout layout = EcosystemManager.getTextLayoutManager().layoutStringSimple(label, SampledTrackGraphViewPort.TIME_HELPER_FONT);
				AxisAlignedBoxBounds rect = layout.getPixelBounds(0, 0);
				int labelWidth  = (int)(rect.getWidth());
				int labelHeight  = (int)(rect.getHeight());
				
				int y = r.getMinY() + (r.getHeight() / 2) - ((labelHeight + 4) / 2);
				
				gm.drawRectangle(new Point(x + 2,  y), new Dimension(labelWidth + 3, labelHeight + 4), 0.0, new Fill(SampledTrackGraphViewPort.TIME_HELPER_COLOR), SampledTrackGraphViewPort.TIME_HELPER_BORDERCOLOR, SampledTrackGraphViewPort.ZOOM_BACKING_BORDER_STROKE, null);

				gm.drawTextLayout(layout, new Point(x + 4, y + labelHeight), Colour.BLACK);
			}
		
		} 
		
//		 Draw Y helper
		if (yAxisTimelineHelperToPaint != null) {
			gm.drawLine(0, yAxisTimelineHelperToPaint.value, EcosystemManager.getGraphicsManager().getWindowSize().getWidth(), yAxisTimelineHelperToPaint.value, Y_HELPER_LINES_COLOR, Y_HELPER_LINES_STROKE);
		}
		
	
	}

	public void paintPreLayeredPanePass()
	{
		if (Browser._theBrowser == null) return;
		
		if (ExpandedTrackManager.getInstance().isAnyExpandedTrackVisible()) {
			Dimension windowSize = EcosystemManager.getGraphicsManager().getWindowSize();
			EcosystemManager.getGraphicsManager().drawRectangle(Point.ORIGIN, windowSize, 0.0, new Fill(SHADED_EXPANDED_BACKING_COLOR), null, null, null);
		}
		
	}

	public Clip paintStarted(Clip currentClip)
	{
		if (Browser._theBrowser == null) return currentClip;

		LinkedList<Widget> seen = null;
		
		//boolean idDirty = false;
		//if (helperLinesToPaint != null) {
		//	idDirty = true;
			
		//}
		
		invalidateHelperLines(currentClip);
		
		timeHelperLinesToPaint = null;
		yAxisTimelineHelperToPaint = null;
		shouldDrawFullscreenHelpers = StandardInputEventListeners.kbmStateListener.isKeyDown(Key.CTRL) || 
			(StandardInputEventListeners.kbmStateListener.isKeyDown(Key.SHIFT) && !FreeItems.getInstance().isEmpty());

		// Get free tracks and paint line helpers
		if (!FreeItems.getInstance().isEmpty() || StandardGestureActions.getlastHighlightedItem() != null) {

			LinkedList<Item> possibles = new LinkedList<Item>(FreeItems.getInstance());
			if (FreeItems.getInstance().isEmpty() && StandardGestureActions.getlastHighlightedItem() != null) {
				possibles.add(StandardGestureActions.getlastHighlightedItem());
			}
			
			for (Item i : possibles) {
				
				Widget iw = null;
				
				if (i instanceof WidgetCorner) {
					iw = ((WidgetCorner)i).getWidgetSource();
				} else if (i instanceof WidgetEdge) {
					iw = ((WidgetEdge)i).getWidgetSource();
				}
				
				if (iw != null) {
					
					if (iw != null && (iw instanceof SampledTrack ||
							iw instanceof LinkedTrack)) {
						
						if (seen == null) {
							seen = new LinkedList<Widget>();
						} else if (seen.contains(iw)) {
							continue;
						}
						
						if (timeHelperLinesToPaint == null)
							timeHelperLinesToPaint = new LinkedList<Integer>();
		
						timeHelperLinesToPaint.add(iw.getX());
					}
				} else if (i instanceof FramePlaybackLauncher || i instanceof RecordOverdubLauncher) {
					if (timeHelperLinesToPaint == null)
						timeHelperLinesToPaint = new LinkedList<Integer>();
					timeHelperLinesToPaint.add(i.getX());
				}
			}

		} 
		
		if ((ApolloGestureActions.isYAxisRestictionOn() || ApolloGestureActions.isSnapOn()) 
			&& !FreeItems.getInstance().isEmpty()) {
			
			int smallestY = FreeItems.getInstance().get(0).getY();
			for (Item i : FreeItems.getInstance()) {
				if (i.getY() < smallestY) smallestY = i.getY();
			}

			if (smallestY > 0) {
				yAxisTimelineHelperToPaint = Mutable.createMutableInteger(smallestY);
			}
		}
	
		// Invalidate on the fly
		invalidateHelperLines(currentClip);
		
		return currentClip;
		
	}
	
	private void invalidateHelperLines(Clip currentClip)
	{
		if (currentClip == null) return;
		
		if (timeHelperLinesToPaint != null && !timeHelperLinesToPaint.isEmpty()) {
			currentClip.includeBounds(OverdubbedFrameTimeAxis.getInstance().getCurrentTimelineArea().clone());
			if (shouldDrawFullscreenHelpers) {
				int fsHeight = EcosystemManager.getGraphicsManager().getWindowSize().getHeight();
				for (Integer n : timeHelperLinesToPaint) {
					currentClip.includeBounds(new AxisAlignedBoxBounds(n - 1, 0, 3, fsHeight));
				}
			}
		}
		
		if (yAxisTimelineHelperToPaint != null && shouldDrawFullscreenHelpers) {

			currentClip.includeBounds(new AxisAlignedBoxBounds(
					0, 
					yAxisTimelineHelperToPaint.value - 1, 
					EcosystemManager.getGraphicsManager().getWindowSize().getWidth(), 
					3));
		
		}
	}
}
