package org.apollo.io;

import java.net.URL;
import java.util.HashMap;

import org.apollo.gui.Strokes;
import org.expeditee.core.Colour;
import org.expeditee.core.Image;
import org.expeditee.core.Line;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;

/**
 * An icon provider. Only loads icons once and when they are requested (simple proxy approach).
 * 
 * @author Brook Novak
 *
 */
public final class IconRepository {
	
	private static HashMap<String, Image> _loadedIcons = new HashMap<String, Image>();
	
	private static Image missingIcon = null;

	private IconRepository()
	{
	}
	
	private static Image getMissingIcon()
	{
		if (missingIcon == null) {
			
			Image image = EcosystemManager.getImageManager().createImage(16, 16, false);
			
			GraphicsManager g = EcosystemManager.getGraphicsManager();
			
			g.pushDrawingSurface(image);
			
			g.clear(Colour.WHITE);

			Line l1 = new Line(0, 0, 16, 16);
			Line l2 = new Line(0, 16, 16, 0);

			g.drawLine(l1, Colour.RED, Strokes.SOLID_2);
			g.drawLine(l2, Colour.RED, Strokes.SOLID_2);
			
			g.popDrawingSurface();
			
			missingIcon = image;
		}
			
		return missingIcon;
	}
	
	/**
	 * 
	 * @param name Exclude path / package
	 * @return
	 */
	public static Image getIcon(String name)
	{
		Image icon = _loadedIcons.get(name);
		
		if (icon == null) {
			try {
				URL url;
				ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
				if (classLoader!=null) {
				    // Applet friendly
				    url = classLoader.getResource("org/apollo/assets/icons/" + name);
				} else {
				    url = ClassLoader.getSystemResource("org/apollo/assets/icons/" + name);
				}

				if (url != null) icon = Image.getImage(url);
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			if (icon == null) { 
				System.err.println("WARNING: Cannot find icon named \"" + name +"\"");
				return getMissingIcon();
			}
		} 
		
		return icon;
	}
	
}
