package org.apollo.items;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.core.Clip;
import org.expeditee.core.Colour;
import org.expeditee.core.Cursor;
import org.expeditee.core.Font;
import org.expeditee.core.Point;
import org.expeditee.core.EnforcedClipStack.EnforcedClipKey;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.gio.gesture.StandardGestureActions;
import org.expeditee.gui.DisplayController;
import org.expeditee.gui.Frame;
import org.expeditee.gui.Popup;
import org.expeditee.gui.PopupManager;
import org.expeditee.items.Text;
import org.expeditee.items.Item.HighlightMode;

/**
 * Its a terrible hack - but it will do.
 * 
 * @author Brook Novak
 *
 */
public class EmulatedTextItem
{
	private List<TextChangeListener> textChangeListeners = new LinkedList<TextChangeListener>();
	
	private RestrictedTextItem emulatedSource;
	private LabelEditPopup popup = null;
	
	private static final Font DEFAULT_FONT = new Font("Serif-Plain", Font.Style.PLAIN, 12);
	
	/**
	 * Constructor.
	 * 
	 * @param parent
	 * 		Must not be null: the parent that lives directly on the content pane.
	 * 
	 * @param offset
	 * 		The offset to add to the label position - from the parents top left corner. 
	 * 		Must not be null.
	 */
	public EmulatedTextItem(String text, Point position)
	{
		assert(position != null);
		
		emulatedSource = new RestrictedTextItem();
		emulatedSource.setPosition(position);
		emulatedSource.setFont(DEFAULT_FONT);
		emulatedSource.setText(text);
		
		popup = new LabelEditPopup();
		PopupManager.getInstance().add(popup);
	}
	
	@Override
	public void finalize()
	{
		PopupManager.getInstance().remove(popup);
	}
	
	public void setFontStyle(Font.Style style)
	{
		emulatedSource.setFont(emulatedSource.getFont().clone().setStyle(style));
	}
	
	public void setFontSize(float size)
	{
		emulatedSource.setSize(size);
	}
	
	public void setPosition(Point position)
	{
		emulatedSource.setPosition(position);
	}
	
	/**
	 * Determines whether or no a mouse event is on the emulated text item.
	 * @param swingME
	 * 		Can be null.
	 * 
	 * @return
	 * 		True if the event is over the emulated item.
	 */
	private boolean isOnEmulatedText()
	{
		Point p = EcosystemManager.getInputManager().getCursorPosition();
		
		return isOnEmulatedText(p.x, p.y);
	}
	
	private boolean isOnEmulatedText(int expX, int expY)
	{
		return emulatedSource.contains(expX, expY);
	}
	
	/**
	 * Applies highlights - shows popup
	 * 
	 * @param e
	 * @param parent
	 */
	public boolean onMouseMoved(MouseEvent e)
	{
		if (isOnEmulatedText()) {
			
			if (emulatedSource.getHighlightMode() != HighlightMode.Normal) {
				emulatedSource.setHighlightMode(HighlightMode.Normal);
				emulatedSource.setHighlightColorToDefault();
				repaint();
			}
			
			return true;
		} else {
			lostFocus();
		}
		
		return false;
	}
	
	public boolean onKeyTyped(KeyEvent e)
	{
		Point p = EcosystemManager.getInputManager().getCursorPosition();

		// If so and the event really comes from its parent ...
		if (isOnEmulatedText(p.x, p.y)) {
			
			int yOffset = -6;
			
			EcosystemManager.getGraphicsManager().setCursor(new Cursor(Cursor.CursorType.TEXT));

			// If a alphanumeric text entry...
			if (!e.isActionKey() &&
					Character.isLetterOrDigit(e.getKeyChar()) || 
					(e.getKeyChar() == Text.BACKSPACE_CHARACTER && !emulatedSource.isEmpty()) ||
					(e.getKeyChar() == Text.DELETE_CHARACTER && !emulatedSource.isEmpty()) ||
					e.getKeyChar() == ' ' ||
					e.getKeyChar() == '_' ||
					e.getKeyChar() == '&' ||
					e.getKeyChar() == '(' ||
					e.getKeyChar() == ')' ||
					e.getKeyChar() == '-'
				)
			{
				// Insert the text according to the current mouse position
				Point newMouse = emulatedSource.insertChar(e.getKeyChar(), p.x, p.y);
				
				List<String> lines = emulatedSource.getTextList();
				if (lines != null && lines.size() > 1) {
					emulatedSource.setText(lines.get(0));
				}
				
				// Notify observers
				fireTextChanged();

				// Move "cursored mouse"
				if (newMouse != null) {
					EcosystemManager.getInputManager().setCursorPosition(new Point((int) newMouse.x, (int) newMouse.y + yOffset));
				}
			}
			
			if (emulatedSource.hasSelection()) {
				emulatedSource.clearSelection();
				repaint();
			}
			
			return true;
			
		}
		
		return false;
	}
	
	/**
	 * 
	 * @param e
	 * 
	 * @param parent
	 */
	public boolean onKeyPressed(KeyEvent e)
	{
		Point p = EcosystemManager.getInputManager().getCursorPosition();

		// If so and the event really comes from its parent ...
		if (isOnEmulatedText(p.x, p.y)) {
			
			int yOffset = -6;
			
			EcosystemManager.getGraphicsManager().setCursor(new Cursor(Cursor.CursorType.TEXT));

			// If a alphanumeric text entry...
			if (e.getKeyCode() == KeyEvent.VK_LEFT || e.getKeyCode() == KeyEvent.VK_RIGHT) {
				
				Point newMouse = emulatedSource.moveCursor(
						(e.getKeyCode() == KeyEvent.VK_LEFT) ? Text.LEFT : Text.RIGHT, 
						(float)p.getX(), (float)p.getY(), false, false);
				
				if (newMouse != null) {
					EcosystemManager.getInputManager().setCursorPosition(new Point((int) newMouse.x, (int) newMouse.y + yOffset));
				}

			}
			
			if (emulatedSource.hasSelection()) {
				emulatedSource.clearSelection();
				repaint();
			}
			
			return true;
			
		}
		
		return false;
	}
	
	/**
	 * 
	 * @param e
	 * 
	 * @param parent
	 */
	public boolean onKeyReleased(KeyEvent e)
	{
		return isOnEmulatedText();
	}
	
	
	public boolean onMouseReleased(MouseEvent e)
	{
		Point p = EcosystemManager.getInputManager().getCursorPosition();

		// If so and the event really comes from its parent ...
		if (isOnEmulatedText(p.x, p.y)) {
			
			String toMoveIntoFreespace = null;
			
			if (e.getButton() == MouseEvent.BUTTON1) {
				// LINK
			} else if (e.getButton() == MouseEvent.BUTTON2 && emulatedSource.hasSelection() && !emulatedSource.isEmpty()) {
	
				invalidate();
				
				toMoveIntoFreespace = emulatedSource.cutSelectedText();
				
			} else if (e.getButton() == MouseEvent.BUTTON3 && emulatedSource.hasSelection() && !emulatedSource.isEmpty()) {
				toMoveIntoFreespace = emulatedSource.copySelectedText();
				
			} else if (e.getButton() == MouseEvent.BUTTON3 && 
					!emulatedSource.hasSelection() && 
					!emulatedSource.isEmpty()) {
				toMoveIntoFreespace = emulatedSource.getText();
			}
			
			if (toMoveIntoFreespace != null) {
				
				Frame target = DisplayController.getCurrentFrame();
				if (target != null) {
					Text selectionCopy = new Text(target.getNextItemID(), toMoveIntoFreespace);
					selectionCopy.setPosition(p.x, p.y);
					selectionCopy.setSize(emulatedSource.getSize());
					StandardGestureActions.pickup(selectionCopy);
					lostFocus();
				}
				
				
			}
			
			emulatedSource.clearSelection();
			repaint();
			
			return true;
		}
		
		return false;
	}
	
	public boolean onMousePressed(MouseEvent e)
	{
		Point p = EcosystemManager.getInputManager().getCursorPosition();

		// If so and the event really comes from its parent ...
		if (isOnEmulatedText(p.x, p.y)) {
			
			emulatedSource.clearSelection();
			emulatedSource.setSelectionStart(p.x, p.y, e.getButton());

			repaint();
			
			return true;
		}
		
		return false;
	}
	
	public boolean onMouseClicked(MouseEvent e)
	{
		return isOnEmulatedText();
	}
	
	public boolean onMouseDragged(MouseEvent e)
	{
		Point p = EcosystemManager.getInputManager().getCursorPosition();

		// If so and the event really comes from its parent ...
		if (isOnEmulatedText(p.x, p.y)) {

			EcosystemManager.getGraphicsManager().setCursor(new Cursor(Cursor.CursorType.TEXT));
			
			emulatedSource.setSelectionEnd(p.x, p.y);
			repaint();
			
			return true;
		}
		
		return false;
	}
	
	public void addTextChangeListener(TextChangeListener listener)
	{
		assert(listener != null);
		if (!textChangeListeners.contains(listener)) textChangeListeners.add(listener);
	}
	
	public void gainFocus()
	{
		if (popup != null && !PopupManager.getInstance().isShowing(popup)) {
			popup.show();
		}
	}
	
	public void lostFocus()
	{
		EcosystemManager.getGraphicsManager().setCursor(new Cursor(Cursor.CursorType.DEFAULT));
		
		boolean dirty = false;
		if (emulatedSource.hasSelection()) {
			emulatedSource.clearSelection();
			dirty = true;
		}
		
		if (emulatedSource.getHighlightMode() != HighlightMode.None) {
			emulatedSource.setHighlightMode(HighlightMode.None);
			emulatedSource.setHighlightColorToDefault();
			dirty = true;
		}

		if (popup != null && PopupManager.getInstance().isShowing(popup)) {
			popup.hide();
		}
		
		if (dirty) repaint();
		
	}
	
	/**
	 * Sets the text and fires a text changed event if the text has changed.
	 * 
	 * @param text
	 * 		Null is allowed
	 */
	public void setText(String text)
	{
		if (text == null) text = "";
		
		if (emulatedSource.getText().equals(text)) return;
		
		invalidate();
		
		this.emulatedSource.setText(text);
		
		invalidate();
		
		fireTextChanged();
	}
	
	public String getText()
	{
		return emulatedSource.getText();
	}
	
	private void fireTextChanged()
	{
		for (TextChangeListener listener : textChangeListeners) {
			listener.onTextChanged(this, this.emulatedSource.getText());
		}
	}

	public void setBackgroundColor(Colour c)
	{
		emulatedSource.setBackgroundColor(c);
	}
	
	private void repaint()
	{
		if (popup != null && PopupManager.getInstance().isShowing(popup)) {
			popup.invalidateAppearance();
			DisplayController.requestRefresh(true);
		} else {
			invalidate();
			DisplayController.requestRefresh(true);
		}
	}
	
	private void invalidate()
	{
		DisplayController.invalidateArea(emulatedSource.getDrawingArea());
	}
	
	public void paint()
	{
		GraphicsManager gm = EcosystemManager.getGraphicsManager();

		EnforcedClipKey key = gm.pushClip(new Clip(emulatedSource.getDrawingArea()));
		
		emulatedSource.paint();

		gm.popClip(key);
	}
	
	public interface TextChangeListener
	{
		public void onTextChanged(Object source, String newLabel);
	}
	
	private class LabelEditPopup extends Popup implements KeyListener, MouseListener, MouseMotionListener
	{
		LabelEditPopup()
		{
			super(null); // No animation (instant show/hide)
			super.setConsumeBackClick(false);
			super.setBorderThickness(0);
		}

		@Override
		public void onHide()
		{
		}
	
		@Override
		public void onShow()
		{
		}
		
		@Override
		public void paintInternal()
		{
			EmulatedTextItem.this.paint();
		}

		@Override
		public AxisAlignedBoxBounds getFullBounds()
		{
			return emulatedSource.getDrawingArea();
		}

		@Override
		public void mouseDragged(MouseEvent e)
		{
			boolean consume = onMouseDragged(e);
			if (consume) e.consume();
		}

		@Override
		public void mouseMoved(MouseEvent e)
		{
			boolean consume = onMouseMoved(e);
			if (consume) e.consume();
		}

		@Override
		public void mouseClicked(MouseEvent e)
		{
			boolean consume = onMouseClicked(e);
			if (consume) e.consume();
		}

		@Override
		public void mousePressed(MouseEvent e)
		{
			boolean consume = onMousePressed(e);
			if (consume) e.consume();
		}

		@Override
		public void mouseReleased(MouseEvent e)
		{
			boolean consume = onMouseReleased(e);
			if (consume) e.consume();
		}

		@Override
		public void mouseEntered(MouseEvent e)
		{
			// Not required
		}

		@Override
		public void mouseExited(MouseEvent e)
		{
			// Not required
		}

		@Override
		public void keyTyped(KeyEvent e)
		{
			boolean consume = onKeyTyped(e);
			if (consume) e.consume();
		}

		@Override
		public void keyPressed(KeyEvent e)
		{
			boolean consume = onKeyPressed(e);
			if (consume) e.consume();
		}

		@Override
		public void keyReleased(KeyEvent e)
		{
			boolean consume = onKeyReleased(e);
			if (consume) e.consume();
		}
	
	}
	
	private class RestrictedTextItem extends Text
	{
		private Colour selectionColor = null;
		
		RestrictedTextItem()
		{
			super(1);
		}
		
		private void updateSelection()
		{
			selectionColor = super.getSelectionColour();
		}

		public void setSelectionStart(float mouseX, float mouseY, int mouseButton)
		{
			super.setSelectionStart(mouseX, mouseY);
			updateSelection();
		}
		
		@Override
		public Colour getSelectionColour()
		{
			if (selectionColor != null) {
				return selectionColor;
			} else {
				return super.getSelectionColour();
			}

		}
		
		
	}
}
