/**
 *    XItem.java
 *    Copyright (C) 2010 New Zealand Digital Library, http://expeditee.org
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.expeditee.io.flowlayout;


import org.expeditee.core.Point;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.items.Item;

public abstract class XItem 
{
	protected AxisAlignedBoxBounds bounding_rect;
		
	protected XItem()
	{
		bounding_rect = null;
	}
	
	protected XItem(AxisAlignedBoxBounds imprint_bounding_rect)
	{
		this.bounding_rect = imprint_bounding_rect;
	}
	
	public int getX()
	{
		return bounding_rect.getMinX();
	}
	
	public int getBoundingXLeft()
	{
		// name alias for getX()
		return bounding_rect.getMinX();
	}
	
	public int getY()
	{
		return bounding_rect.getMinY();
	}
	
	public int getBoundingYTop()
	{
		// name alias for getY()
		return bounding_rect.getMinY();
	}
	
	public int getBoundingYBot()
	{
		return bounding_rect.getMaxY();
	}
	
	public int getBoundingXRight()
	{
		return bounding_rect.getMaxX();
	}
	
	public int getBoundingWidth()
	{
		return bounding_rect.getWidth();
	}
	
	public int getBoundingHeight()
	{
		return bounding_rect.getHeight();
	}
	
	public AxisAlignedBoxBounds getBoundingRect()
	{
		return bounding_rect;
	}
	
	public AxisAlignedBoxBounds setBoundingRect(AxisAlignedBoxBounds rect)
	{
		return bounding_rect = rect;
	}
	
	public boolean containsItem(Item item)
	{
		int x = item.getX();
		int y = item.getY();
		
		Point pt = new Point(x,y);
		
		return bounding_rect.contains(pt);
	}
	
	public boolean containedInXExtent(int x) 
	{
		int xl = getBoundingXLeft();
		int xr = getBoundingXRight();
		
		return (x>= xl) && (x<=xr);
	}
	
	
	public boolean containedInYExtent(int y) 
	{
		int yt = getBoundingYTop();
		int yb = getBoundingYBot();
		
		return (y>= yt) && (y<=yb);
	}
	
	
}
