/**
 *    WidgetEdge.java
 *    Copyright (C) 2010 New Zealand Digital Library, http://expeditee.org
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.expeditee.items.widgets;

import java.util.List;

import org.expeditee.core.Colour;
import org.expeditee.core.Point;
import org.expeditee.core.bounds.PolygonBounds;
import org.expeditee.items.Item;
import org.expeditee.items.Line;
import org.expeditee.items.Text;

/**
 * Widget edges define the boundaries of an interactive widget.
 * 
 * @author Brook Novak
 */
public class WidgetEdge extends Line {

	private Widget _widgetSource;

	WidgetEdge(WidgetCorner start, WidgetCorner end, int id,
			Widget widgetSource) {
		super(start, end, id);
		
		if (widgetSource == null)
			throw new NullPointerException("widgetSource");
		_widgetSource = widgetSource;
	}

	public Widget getWidgetSource() {
		return _widgetSource;
	}

	@Override
	public Item forceMerge(Item merger, int mouseX, int mouseY) {
		return null; // Don't allow breaking up of borders
	}

	@Override
	public void toggleArrow() {
		// Ignore
	}

	@Override
	public void toggleDashed(int amount) {
		// Ignore
	}

	@Override
	public void toggleArrowHeadRatio(int amount) {
		// Ignore
	}

	@Override
	public void toggleArrowHeadLength(int amount) {
		// Ignore
	}

	@Override
	public void setAnnotation(boolean val) {
		// Ignore
	}

	@Override
	public boolean isAnnotation() {
		return false;
	}

	@Override
	public void setArrow(float length, double ratio, double nib_perc) {
	}

	@Override
	public void setArrowhead(PolygonBounds arrow) {
	}

	@Override
	public void setArrowheadLength(float length) {
	}

	@Override
	public void setArrowheadRatio(double ratio) {
	}

	@Override
	public void setBottomShadowColor(Colour bottom) {
	}

	@Override
	public void setFillColor(Colour c) {
	}

	@Override
	public void setFillPattern(String patternLink) {
	}

	@Override
	public void setSize(float size) {
	}
	
	@Override
	public void setFormula(String formula) {
	}
	
	@Override
	public void setAnchorTop(Integer anchor) {
		_widgetSource.setAnchorTop(anchor);
	}

	@Override
	public void setAnchorBottom(Integer anchor) {
		_widgetSource.setAnchorBottom(anchor);
	}

	@Override
	public void setAnchorLeft(Integer anchor) {
		_widgetSource.setAnchorLeft(anchor);
	}

	@Override
	public void setAnchorRight(Integer anchor) {
		_widgetSource.setAnchorRight(anchor);
	}

	@Override
	public boolean isAnchored() {
		return _widgetSource.isAnchored();
	}

	@Override
	public boolean isAnchoredX() {
		return _widgetSource.isAnchoredX();
	}

	@Override
	public boolean isAnchoredY() {
		return _widgetSource.isAnchoredY();
	}
	
	@Override
	public Integer getAnchorTop() {
		return _widgetSource.getSource().getAnchorTop();
	}

	@Override
	public Integer getAnchorBottom() {
		return _widgetSource.getSource().getAnchorBottom();
	}
	
	@Override
	public Integer getAnchorLeft() {
		return _widgetSource.getSource().getAnchorLeft();
	}

	@Override
	public Integer getAnchorRight() {
		return _widgetSource.getSource().getAnchorRight();
	}
	
	@Override
	public boolean contains(Point p) {
		return super.contains(p) && !_widgetSource.getBounds().contains(p);
	}

	@Override
	public PolygonBounds getEnclosedShape() {
		return getStartItem().getEnclosedShape();
	}

	@Override
	public String getLink() {
		return _widgetSource.getSource().getLink();
	}

	@Override
	public void setLink(String link) {
		_widgetSource.setLink(link, null);
	}
	
	/**
	 * @param link
	 *          The new frame link. Can be null (for no link)
	 * 
	 * @param linker
	 * 			The text item creating the link. Null if not created from
	 * 			a text item.
	 */
	public void setLink(String link, Text linker) {
		_widgetSource.setLink(link, linker);
	}

	@Override
	public void setThickness(float newThickness, boolean setConnected) {
		if (_widgetSource != null) {
			float minThickness = _widgetSource.getMinimumBorderThickness();
			if(newThickness < minThickness)
				newThickness = minThickness;
			super.setThickness(newThickness, setConnected);
			_widgetSource.setSourceThickness(newThickness, false);
		}
	}	
	
	@Override
	public void setBackgroundColor(Colour c) {
		if (_widgetSource != null) {
			super.setBackgroundColor(c);
			_widgetSource.setBackgroundColor(c);
		}
	}
	
	@Override
	public void setColor(Colour c) {
		if (_widgetSource != null) {
			super.setColor(c);
			_widgetSource.setSourceBorderColor(c);
		}
	}
	
	@Override
	public void setData(List<String> data) {
		_widgetSource.setSourceData(data);
	}
	
	@Override
	public String getName() {
		return _widgetSource.getName();
	}
	
}
